/*
 *  apcaccess.c  -- Text based IPC management tool for apcupsd package.
 *
 *  apcupsd.c -- Simple Daemon to catch power failure signals from a
 *               BackUPS, BackUPS Pro, or SmartUPS (from APCC).
 *            -- Now SmartMode support for SmartUPS and BackUPS Pro.
 *
 *  Copyright (C) 1996-98 Andre M. Hedrick <hedrick@astro.dyer.vanderbilt.edu>
 *                Vanderbilt University, Dyer Observatory.
 *  All rights reserved.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <ctype.h>
#include <pwd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <apc_defines.h>
#include <apc_struct.h>
#include <apc_extern.h>

#define BUFFER_SIZE PIPE_BUF

struct  passwd  *pwentry;

int server_fifo_fd;
int client_fifo_fd;

char msg[100];
time_t nowtime;
char m_char;

UPSINFO myUPS;

DATAINFO myDATA =
{ "\0",          /* data->apcmagic             */
  0,             /* data->update_master_config */
  0,             /* data->get_master_status    */
  -1,            /* data->slave_status         */
  0,             /* data->call_master_shutdown */
  "\0"           /* data->accessmagic          */
};

CONFIGINFO myCONFIG =
{ 0,             /* config->new_annoy          */
  0,             /* config->new_maxtime        */
  0,             /* config->new_delay          */
  0,             /* config->new_proctime       */
  0,             /* config->new_logtime        */
  0,             /* config->new_nettime        */
  0,             /* config->new_percent        */

  "\0",          /* config->new_batt_cmd       */
  "\0",          /* config->new_time_cmd       */
  "\0",          /* config->new_load_cmd       */
  "\0",          /* config->new_pwr_cmd        */
  "\0",          /* config->new_ret_cmd        */
  "\0"           /* config->new_remote_cmd     */
};

/*********************************************************************/
void output_status(UPSINFO *ups);
void reconfig_data(CONFIGINFO *config);
void do_main_menu(void);
void do_config_menu(void);

/*********************************************************************/
void output_status(UPSINFO *ups)
{ time(&nowtime);
  strftime(msg, 100, "%b %d %X", localtime(&nowtime));

  fprintf(stderr, "\nAPC      : %s", msg);
  fprintf(stderr, "\nCABLE    : %s", ups->cable.long_name);
  fprintf(stderr, "\nUPSMODEL : %s", ups->mode.long_name);
  fprintf(stderr, "\nUPSMODE  : %s", ups->class.long_name);
  if (ups->sharenet.type != DISABLE)
    fprintf(stderr, "\nSHARE    : %s", ups->sharenet.long_name);

  switch(ups->mode.type)
    { case BK:
      case SHAREBASIC:
      case NETUPS:
        if (ups->LineUp == 0)
            { fprintf(stderr, "\nLINEFAIL : OK");
              fprintf(stderr, "\nBATTSTAT : OK\n"); }
          else
            { fprintf(stderr, "\nLINEFAIL : DOWN");
              if (ups->BattUp == 0)
                  fprintf(stderr, "\nBATTSTAT : RUNNING\n");
                else
                  fprintf(stderr, "\nBATTSTAT : FAILING\n"); }
        break;
      case BKPRO:
      case VS:
      case NBKPRO:
      case SMART:
      case SHARESMART:
      case MATRIX:

        if (ups->UPSlink_flags == 0x0000800f)
          { if (ups->LineUp == 0)
                { fprintf(stderr, "\nLINEFAIL : OK");
                  fprintf(stderr, "\nBATTSTAT : OK");
         	        if (ups->LineLevel == -1)
                      fprintf(stderr, "\nLINEVOLT : LOW");
                    else if (ups->LineLevel == 1)
                      fprintf(stderr, "\nLINEVOLT : HIGH");
                    else
                      fprintf(stderr, "\nLINEVOLT : OK"); }
              else
                { fprintf(stderr, "\nLINEFAIL : DOWN");
                  if (ups->BattUp == 0)
                      fprintf(stderr, "\nBATTSTAT : RUNNING");
                    else
                      fprintf(stderr, "\nBATTSTAT : FAILING"); } }

        if (ups->UPSlink_flags & 0x01000000)
          fprintf(stderr, "\nUPSNAME  : %s", ups->name);

        if (ups->UPSlink_flags & 0x00000010)
          fprintf(stderr, "\nULINE    : %05.1f Volts", ups->LineVoltage);

        if (ups->UPSlink_flags & 0x00000020)
          fprintf(stderr, "\nMLINE    : %05.1f Volts", ups->LineMax);

        if (ups->UPSlink_flags & 0x00000040)
          fprintf(stderr, "\nNLINE    : %05.1f Volts", ups->LineMin);

        if (ups->UPSlink_flags & 0x00000800)
          fprintf(stderr, "\nFLINE    : %03.1f Hz", ups->LineFreq);

        if (ups->UPSlink_flags & 0x00000080)
          fprintf(stderr, "\nVOUTP    : %05.1f Volts", ups->OutputVoltage);

        if (ups->UPSlink_flags & 0x00000400)
          fprintf(stderr, "\nLOUTP    : %05.1f Load Capacity", ups->UPSLoad);

        if (ups->UPSlink_flags & 0x00000200)
          fprintf(stderr, "\nBOUTP    : %04.1f Volts", ups->BattVoltage);

        if (ups->UPSlink_flags & 0x00000100)
          fprintf(stderr, "\nBCHAR    : %05.1f Batt. Charge", ups->BatLoad);

        if (ups->UPSlink_flags & 0x00001000)
          fprintf(stderr, "\nTIME     : %04.1f Minutes", ups->TimeLeft);

        if (ups->UPSlink_flags & 0x00010000)
          { switch((*ups).sensitivity[0])
              { case 'L': fprintf(stderr, "\nSENSE    : LOW");break;
                case 'M': fprintf(stderr, "\nSENSE    : MEDIUM");break;
                case 'H':
                default : fprintf(stderr, "\nSENSE    : HIGH");break; } }

        if (ups->UPSlink_flags & 0x00020000)
          fprintf(stderr, "\nWAKEUP   : %03d Cycles", ups->wakeup);

        if (ups->UPSlink_flags & 0x00040000)
          fprintf(stderr, "\nSLEEP    : %03d Cycles", ups->sleep);

        if (ups->UPSlink_flags & 0x00080000)
          fprintf(stderr, "\nLOTRANS  : %03d.0 Volts", ups->lotrans);

        if (ups->UPSlink_flags & 0x00100000)
          fprintf(stderr, "\nHITRANS  : %03d.0 Volts", ups->hitrans);

        if (ups->UPSlink_flags & 0x00200000)
          fprintf(stderr, "\nCHARGE   : %03d.0 Percent", ups->minon);

        if (ups->UPSlink_flags & 0x00000001)
          fprintf(stderr, "\nBFAIL    : 0x%02X Status Flag", ups->Status);

        if (ups->UPSlink_flags & 0x00002000)
          fprintf(stderr, "\nUTEMP    : %04.1f C Internal", ups->UPSTemp);

        if (ups->UPSlink_flags & 0x00400000)
          { switch((*ups).beepstate[0])
              { case 'T': fprintf(stderr, "\nALARM    : Low Batt + 30");break;
                case 'L': fprintf(stderr, "\nALARM    : Low Batt");break;
                case 'N': fprintf(stderr, "\nALARM    : Never");break;
                case '0':
                default : fprintf(stderr, "\nALARM    : Always");break; } }

        if (ups->UPSlink_flags & 0x00000004)
          { switch(ups->G[0])
              { case 'O':
                  fprintf(stderr, "\nLASTEVNT : POWER UP");
                  break;
                case 'S':
                  fprintf(stderr, "\nLASTEVNT : SELF TEST");
                  break;
                case 'L':
                  fprintf(stderr, "\nLASTEVNT : LINE VOLTAGE DECREASE");
                  break;
                case 'H':
                  fprintf(stderr, "\nLASTEVNT : LINE VOLTAGE INCREASE");
                  break;
                case 'T':
                  fprintf(stderr, "\nLASTEVNT : POWER FAILURE");
                  break;
                default :
                  fprintf(stderr, "\nLASTEVNT : UNKNOWN EVENT %c", ups->G[1]);
                  break; } }

        if (ups->UPSlink_flags & 0x00800000)
          fprintf(stderr, "\nLOWBATT  : %02d Minutes", ups->lowbatt);

        if (ups->UPSlink_flags & 0x00004000)
          { if (ups->seven & 0x08) fprintf(stderr, "\nDIPSW    : 0x%d", TRUE);
              else fprintf(stderr, "\nDIPSW    : 0x%d", FALSE);
            if (ups->seven & 0x04) fprintf(stderr, "%d", TRUE);
              else fprintf(stderr, "%d", FALSE);
            if (ups->seven & 0x02) fprintf(stderr, "%d", TRUE);
              else fprintf(stderr, "%d", FALSE);
            if (ups->seven & 0x01) fprintf(stderr, "%d", TRUE);
              else fprintf(stderr, "%d", FALSE); }

        fprintf(stderr, "\n");
        break;
      default:
        break; } }

/*********************************************************************/
void reconfig_data(CONFIGINFO *config)
{ /* config->new_annoy          */
  /* config->new_maxtime        */
  /* config->new_delay          */
  /* config->new_proctime       */
  /* config->new_logtime        */
  /* config->new_nettime        */
  /* config->new_percent        */
  /* config->new_batt_cmd       */
  /* config->new_time_cmd       */
  /* config->new_load_cmd       */
  /* config->new_pwr_cmd        */
  /* config->new_ret_cmd        */
  /* config->new_remote_cmd     */
}

/*********************************************************************/
void do_main_menu(void)
{ char ch, temp_comment[256];
  int comment_count = 0;

  while((ch = getchar()) != '\n')
    temp_comment[comment_count++] = ch;
  temp_comment[comment_count] = '\0';
  m_char = temp_comment[0];
  printf("\n");

  switch(m_char)
    { case '~':
        m_char = 'S';
        break;
      case 'R':
      case 'r':
        do_config_menu();
        break;
      case 'S':
      case 's':
        break;
      case 'K':
      case 'k':
        break;
      case 'E':
      case 'e':
        m_char = '~';
        break;
      default:
        break; }

  system("clear"); }
 
/*********************************************************************/
void do_config_menu(void)
{ system("clear"); }

/*********************************************************************/
int main(int argc, char **argv)
{ int bytes_read = 0;
  int bytes_sent = 0;
  int test = 0;
  int mode = 0;
  int step = 0;

  zero_ups_struct(&myUPS);

  pwentry = getpwuid(getuid());

  if (strcmp("root", pwentry->pw_name) != 0)
    { fprintf(stderr, "%s: YOU MUST BE ROOT TO USE THIS PROGRAM\n", argv[0]);
      fprintf(stderr, "%s: YOU ARE NOT ALLOWED HERE <%s>, GO AWAY!!\n", argv[0], pwentry->pw_name);
      fprintf(stderr, "%s: TRYING TO LOG YOU OFF SYSTEM\n", argv[0]);
      (void) system("logout");
      exit(1); }

  if (argc < 2)
    { fprintf(stderr, "Usage: %s < reconfig | status | slave | shutdown >\n", argv[0]);
      exit(1); }

  if (strcmp(argv[1], "reconfig") == 0)
      { mode = 1;
        fprintf(stderr, "%s <%s> : is not functional yet.\n", argv[0], argv[1]);
        myDATA.update_master_config = TRUE;
        myDATA.get_master_status    = FALSE;
        myDATA.slave_status         = -1;
        myDATA.call_master_shutdown = FALSE;
        exit(1); }
    else if (strcmp(argv[1], "status") == 0)
      { mode = 2;
        myDATA.update_master_config = FALSE;
        myDATA.get_master_status    = TRUE;
        myDATA.slave_status         = -1;
        myDATA.call_master_shutdown = FALSE;
        fprintf(stderr, "%s : polling apcupsd for %s.\n", argv[0], argv[1]); }
    else if (strcmp(argv[1], "slave") == 0)
      { mode = 3;
        myDATA.update_master_config = FALSE;
        myDATA.get_master_status    = FALSE;
        myDATA.slave_status         = TRUE;
        myDATA.call_master_shutdown = FALSE;
        fprintf(stderr, "%s : calling apcupsd for %s\n", argv[0], argv[1]); }
    else if (strcmp(argv[1], "shutdown") == 0)
      { mode = 4;
        myDATA.update_master_config = FALSE;
        myDATA.get_master_status    = FALSE;
        myDATA.slave_status         = -1;
        myDATA.call_master_shutdown = TRUE;
        fprintf(stderr, "%s : calling apcupsd for %s.\n", argv[0], argv[1]); }
    else
      { mode = 0;
        exit(1); }

/*
  while(m_char != '~') do_main_menu();
*/

  mkfifo(ACCESS_FIFO, 0777);
  if ((server_fifo_fd = open(ACCESS_FIFO, O_WRONLY)) == -1)
    { fprintf(stderr, "Server fifo failure\n");
      unlink(ACCESS_FIFO);
      exit(1); }

  strcpy(myDATA.apcmagic, APC_MAGIC);
  strcpy(myDATA.accessmagic, ACCESS_MAGIC);

  for(step=0;step<12;step++)
    { if ((client_fifo_fd = open(PIPE_FIFO, O_RDONLY)) != -1)
        goto fifo_success;
      sleep(5); }


  close(client_fifo_fd);
  close(server_fifo_fd);
  fprintf(stderr, "client fifo failure\n");
  unlink(ACCESS_FIFO);
  exit(1);

fifo_success:

  switch(mode)
    { case 1:
        reconfig_data(&myCONFIG);
        if ((test = write(server_fifo_fd, &myDATA, sizeof(myDATA))) != sizeof(myDATA))
          { fprintf(stderr, "\n%s: %d size of myDATA", argv[0], test);
            close(server_fifo_fd);
            close(client_fifo_fd);
            unlink(ACCESS_FIFO);
            exit(1); }
        while(bytes_sent < sizeof(myCONFIG))
          { test = write(server_fifo_fd, &myCONFIG, BUFFER_SIZE);
            bytes_sent += test; }
        close(server_fifo_fd);
        if ((test = read(client_fifo_fd, &myDATA, sizeof(myDATA))) != sizeof(myDATA))
          { fprintf(stderr, "\n%s: %d size of myDATA", argv[0], test);
            close(client_fifo_fd);
            unlink(ACCESS_FIFO);
            exit(1); }
        close(client_fifo_fd);
        break;
      case 2:
        if ((test = write(server_fifo_fd, &myDATA, sizeof(myDATA))) != sizeof(myDATA))
          { fprintf(stderr, "\n%s: %d size of myDATA", argv[0], test);
            close(server_fifo_fd);
            close(client_fifo_fd);
            unlink(ACCESS_FIFO);
            exit(1); }
        do { test = read(client_fifo_fd, &myUPS, BUFFER_SIZE);
             bytes_read += test; } while (test > 0) ;
        close(client_fifo_fd);
        output_status(&myUPS);
        break;
      case 3:
      case 4:
        if ((test = write(server_fifo_fd, &myDATA, sizeof(myDATA))) != sizeof(myDATA))
          { fprintf(stderr, "\n%s: %d size of myDATA", argv[0], test);
            close(server_fifo_fd);
            close(client_fifo_fd);
            unlink(ACCESS_FIFO);
            exit(1); }
        close(client_fifo_fd);
        break; }

  close(server_fifo_fd);
  unlink(ACCESS_FIFO);

  return(0); }
