/*
 *  apcconfig.c  -- configuration parts for apcupsd package
 *
 *  apcupsd.c    -- Simple Daemon to catch power failure signals from a
 *                  BackUPS, BackUPS Pro, or SmartUPS (from APCC).
 *               -- Now SmartMode support for SmartUPS and BackUPS Pro.
 *
 *  Copyright (C) 1996-98 Andre M. Hedrick <hedrick@astro.dyer.vanderbilt.edu>
 *                Vanderbilt University, Dyer Observatory.
 *  All rights reserved.
 *
 * Heavily modified by:
 *     Riccardo Facchetti <fizban@tin.it>
 *      Jonathan H N Chin <jc254@newton.cam.ac.uk>
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdlib.h>
#include <errno.h>

#include "apc_defines.h"
#include "apc_struct.h"
#include "apc_extern.h"

/* ---------------------------------------------------------------------- */

#define WHERE(MEMBER) ((size_t) &((UPSINFO *)0)->MEMBER)
#define AT(UPS,OFFSET) ((size_t)UPS + OFFSET)

typedef int (HANDLER)(UPSINFO *, size_t, GENINFO *, const char *);

static HANDLER match_cmd, match_int, match_log, match_range,
               match_slave, match_str;

/* ---------------------------------------------------------------------- */

static GENINFO accesses[] = {
	{ "false",  "false",                 FALSE }, /* must come first */
	{ "true",   "true",                  TRUE },
	{ NULL,     "*invalid-access-mode*", FALSE },
};

static GENINFO cables[] = {
	{ "simple",    "Custom Cable Simple", CUSTOM_SIMPLE },
	{ "smart",     "Custom Cable Smart",  CUSTOM_SMART },
	{ "ether",     "Ethernet Link",       APC_NET },
	{ "940-0020B", "APC Cable 940-0020B", APC_940_0020B },
	{ "940-0023A", "APC Cable 940-0023A", APC_940_0023A },
	{ "940-0024B", "APC Cable 940-0024B", APC_940_0024B },
	{ "940-0024C", "APC Cable 940-0024C", APC_940_0024C },
	{ "940-0095A", "APC Cable 940-0095A", APC_940_0095A },
	{ "940-0095C", "APC Cable 940-0095C", APC_940_0095C },
	{ NULL,        "*invalid-cable*",     NO_CABLE  },
};

static GENINFO classes[] = {
	{ "standalone",     "Stand Alone",           STANDALONE },
	{ "shareslave",     "ShareUPS Slave",        SHARESLAVE },
	{ "sharemaster",    "ShareUPS Master",       SHAREMASTER },
	{ "netslave",       "Net Slave",             NETSLAVE },
	{ "netmaster",      "Net Master",            NETMASTER },
	{ "sharenetmaster", "ShareUPS & Net Master", SHARENETMASTER },
	{ NULL,             "*invalid-class-mode*",  NO_CLASS },
};

static GENINFO logins[] = {
	{ "always",  "always",               ALWAYS }, /* must come first */
	{ "disable", "disable",              NEVER },
	{ "timeout", "timeout",              TIMEOUT },
	{ "percent", "percent",              PERCENT },
	{ "minutes", "minutes",              MINUTES },
	{ NULL,      "*invalid-login-mode*", NO_LOGON },
};

static GENINFO modes[] = {
	{ "disable",  "Network & ShareUPS Disabled", DISABLE },
	{ "share",    "ShareUPS",                    SHARE },
	{ "net",      "NetworkUPS",                  NET },
	{ "sharenet", "Network & ShareUPS",          SHARENET },
	{ NULL,       "*invalid-sharenet-mode*",     NO_SHARE_NET },
};

static GENINFO types[] = {
	{ "backups",       "BackUPS",                   BK },
	{ "sharebasic",    "ShareUPS Basic Port",       SHAREBASIC },
	{ "netups",        "NetUPS Virtual Basic Port", NETUPS },
	{ "backupspro",    "BackUPS Pro",               BKPRO },
	{ "smartvsups",    "SmartUPS VS",               VS },
	{ "newbackupspro", "Smarter BackUPS Pro",       NBKPRO },
	{ "smartups",      "SmartUPS",                  SMART },
	{ "matrixups",     "MatrixUPS",                 MATRIX },
	{ "sharesmart",    "ShareUPS Advanced Port",    SHARESMART },
	{ NULL,            "*invalid-mode*",            NO_UPS },
};

static struct pairs {
	const char *key;
	HANDLER *handler;
	size_t offset;
	GENINFO *values;
} table[] = {

	{ "#",        0, -1, 0 },
	{ "\n",       0, -1, 0 },
	{ "SELFTEST", 0, -1, 0 },

	{ "ANNOY",        match_int, WHERE(annoy),      0 },
	{ "BATTERYLEVEL", match_int, WHERE(percent),    0 },
	{ "DELAY",        match_int, WHERE(delay),      0 },
	{ "HITRANSFER",   match_int, WHERE(hitrans),    0 },
	{ "LOTRANSFER",   match_int, WHERE(lotrans),    0 },
	{ "LOWBATT",      match_int, WHERE(lowbatt),    0 },
	{ "MINUTES",      match_int, WHERE(runtime),    0 },
	{ "NETPORT",      match_int, WHERE(NetUpsPort), 0 },
	{ "NETTIME",      match_int, WHERE(nettime),    0 },
	{ "PROCFS",       match_int, WHERE(proctime),   0 },
	{ "RETURNCHARGE", match_int, WHERE(minon),      0 },
	{ "SLEEP",        match_int, WHERE(sleep),      0 },
	{ "TIMEOUT",      match_int, WHERE(maxtime),    0 },
	{ "WAKEUP",       match_int, WHERE(wakeup),     0 },

	{ "CONTROL",   match_cmd, WHERE(control_path), 0 },
	{ "BATTCMD",   match_cmd, WHERE(batt_cmd),     0 },
	{ "LIMITCMD",  match_cmd, WHERE(limit_cmd),    0 },
	{ "LOADCMD",   match_cmd, WHERE(load_cmd),     0 },
	{ "PWRCMD",    match_cmd, WHERE(pwr_cmd),      0 },
	{ "REMOTECMD", match_cmd, WHERE(remote_cmd),   0 },
	{ "RETCMD",    match_cmd, WHERE(ret_cmd),      0 },
	{ "TIMECMD",   match_cmd, WHERE(time_cmd),     0 },

	{ "LOGGING", match_log,    -1, 0 },
	{ "SLAVE",   match_slave,  -1, 0 },

	{ "BEEPSTATE",   match_str, WHERE(beepstate),   0 },
	{ "DEVICE",      match_str, WHERE(device),      0 },
	{ "LOCKFILE",    match_str, WHERE(lockpath),    0 },
	{ "MASTER",      match_str, WHERE(master_name), 0 },
	{ "SENSITIVITY", match_str, WHERE(sensitivity), 0 },
	{ "UPSNAME",     match_str, WHERE(name),        0 },
	{ "USERMAGIC",   match_str, WHERE(usermagic),   0 },

	{ "ACCESS",   match_range, WHERE(enable_access), accesses },
	{ "NOLOGON",  match_range, WHERE(nologin),       logins },
	{ "UPSCABLE", match_range, WHERE(cable),         cables },
	{ "UPSCLASS", match_range, WHERE(class),         classes },
	{ "UPSMODE",  match_range, WHERE(sharenet),      modes },
	{ "UPSTYPE",  match_range, WHERE(mode),          types },

	/* must be last */
	{ NULL, 0, -1, 0 },
};

/* ---------------------------------------------------------------------- */

/* JHNC-FIXME: any particular reason not to do case-insensitive matching? */
/* JHNC-FIXME: either all match_* can change ups and fail or none should */

/* ---------------------------------------------------------------------- */

static int match_cmd (UPSINFO *ups, size_t offset,
                      GENINFO *junk, const char *v) {
	struct stat statbuf;
	char *x = (char *)AT(ups,offset); /* JHNC-FIXME: x[MAXSTRING] here? */

	if (!sscanf(v, "%s", x))
		return(1);

	/* check existence and exec bit */
	if (!((stat(x, &statbuf) == 0) && (S_ISREG(statbuf.st_mode))
	      && (statbuf.st_mode|S_IEXEC))) {
		fprintf(stderr, "Invalid Command: %s\n", x);
		x[0] = 0; /* JHNC-FIXME: delete this? */
		return(1);
	}
	/* JHNC-FIXME: strcpy here? */
	return(0);
}

static int match_int (UPSINFO *ups, size_t offset,
                      GENINFO *junk, const char *v) {
	int x;
	if (sscanf(v, "%d", &x)) {
		*(int *)AT(ups,offset) = x;
		return(0);
	}
	return(1);
}

/*
 * JHNC-FIXME:
 *
 * Just what is "__NEW_LOGS" anyway?
 * If it doesn't exist, this special case match_log can go away
 * and we just use match_int.
 */
#ifdef __NEW_LOGS
static int match_log (UPSINFO *ups, size_t offset,
                      GENINFO *junk, const char *v) {
	char x[MAXSTRING];
	int time;

	if (!sscanf(v, "%d %s", &time, x))
		return(1);

	strcpy(ups->logfile, x);
	ups->logtime = time;
	return(0);
}
#else
static int match_log (UPSINFO *ups, size_t offset,
                      GENINFO *junk, const char *v) {
	int time;

	if (!sscanf(v, "%d", &time))
		return(1);

	ups->logtime = time;
	return(0);
}
#endif

static int match_range (UPSINFO *ups, size_t offset,
                        GENINFO *vs, const char *v) {
	char x[MAXSTRING];
	INTERNALGENINFO *t;

/* JHNC-FIXME: is this better? seems better than SEGV */
	/* shouldn't ever happen */
	if (!vs) {
		fprintf(stderr,
		        "Bogus configuration table! Fix and recompile.\n");
		exit(1);
	}

	if (!sscanf(v, "%s", x))
		return(1);

	for (; vs->name; vs++)
/*
 * JHNC-FIXME:
 *
 * vs->name is either a statically allocated null-terminated string
 * or someone forgot to give it a proper value (or it is NULL as in
 * the error (last default) case).
 *
 * If it was assigned a value, it must be null-terminated and so there
 * is no point in doing stncmp instead of strcmp.
 *
 * If it is not null-terminated, a strlen check is useless anyway.
 *
 * If this reasoning is right, this comment can be removed, otherwise
 * feel free to change the code back.
 */
		if (!strcmp(v, vs->name))
			break;

	t = (INTERNALGENINFO *)AT(ups,offset);

	/* Copy the structure */
/*
 * JHNC-FIXME:
 *
 * See comment in apcsetup.c : Maybe this can be simplified to the
 * structure assignment I proposed before.
 */
/*
 * JHNC-FIXME:
 *
 * Before the configuration is used, checks could be made for NO_XXX
 * settings. Would mean adding a few lines to check_for_config().
 * Personally, I prefer the "if we have a bogus value, kill the program
 * immediately" approach. In that case, you substitute any `return(1)'
 * and `error++' with an fprintf(stderr, "") and an exit. However.
 *
 * Another approach is:
 *
	if (vs->name) {
		t->type = vs->type;
		strcpy(t->name, vs->name);
		strcpy(t->long_name, vs->long_name);
		return(0);
	} else {
		fprintf(stderr,
		        "Bogus configuration value (%s)\n", vs->long_name);
		return(1);
	}
 */
	if (vs->name)
		strcpy(t->name, vs->name);
	strcpy(t->long_name, vs->long_name);
	t->type = vs->type;

	return(0);
}

static int match_slave (UPSINFO *ups, size_t offset,
                        GENINFO *junk, const char *v) {
	char x[MAXSTRING];

	if (!sscanf(v, "%s", x))
		return(1);

	/* JHNC-FIXME: anything wrong with just using slave_count ? */
	if (ups->slave_count >= MAXSLAVES) {
		fprintf(stderr, "Exceeded max slaves number (%d)\n",
		        MAXSLAVES);
		return(1);
	}
	strcpy(ups->slaves[ups->slave_count].name, x);

	ups->slave_count++;
	return(0);
}

static int match_str (UPSINFO *ups, size_t offset,
                      GENINFO *junk, const char *v) {
	char x[MAXSTRING];

	if (!sscanf(v, "%s", x))
		return(1);

	strcpy((char *)AT(ups,offset), x);
	return(0);
}

/* ---------------------------------------------------------------------- */

static int ParseConfig (UPSINFO *ups, const char *line) {
	struct pairs *p;
	const char *key, *value;

	/* skip initial whitespace */
	for (key = line; *key && (*key==' ' || *key=='\t');) key++;

	/* skip the keyword and whitespace trailing it */
	for (value = key; *value && (*value!=' ' && *value!='\t');) value++;
	for (; *value && (*value==' ' || *value=='\t');) value++;

	if (!*key)
		return(1);

	for (p = table; p->key; p++)
		/*
		 * FIXME
		 *
		 * Here we _must_ use strncmp because the line just read can
		 * be not only a keyword with a specific format (so that
		 * strcmp is enough): it can be a comment ... a string like
		 * "# This is a comment" that we are strcmp() with "#" and
		 * obviously the test must pass even if the strings are
		 * different !
		 *
		 * The line parsing code seems a bit sloppy but for now is
		 * working enough.
		 *
		 * -RF
		 */
		if (!strncmp(key, p->key, strlen(p->key)))
			break;
	if (p->key) {
		if (p->handler)
			return(p->handler(ups, p->offset, p->values, value));
		else
			return(0);
	} else {
		return(1);
	}
}

/* ---------------------------------------------------------------------- */

void zero_ups_struct (UPSINFO *ups) {
	memset(ups, 0, sizeof(UPSINFO));
	strcpy(ups->beepstate,   "0");
	strcpy(ups->sensitivity, "H");
	ups->fd            = -1;

        strcpy(ups->enable_access.name, accesses[0].name);
        strcpy(ups->enable_access.long_name, accesses[0].long_name);
        ups->enable_access.type = accesses[0].type;

        strcpy(ups->nologin.name, logins[0].name);
        strcpy(ups->nologin.long_name, logins[0].long_name);
        ups->nologin.type = logins[0].type;

	ups->annoy         = 60;
	ups->nologin_file  = FALSE;
	ups->nettime       = 60;
	ups->percent       = 10;
	ups->runtime       = 5;
	ups->sleep         = 20;
	ups->lotrans       = 2;
	ups->hitrans       = 2;
	ups->minon         = 3;
	ups->lowbatt       = 2;
	ups->lockfile      = -1;
	ups->remote_state  = TRUE;
}

int check_for_config (UPSINFO *ups)
{
	FILE *apcconf;
	char line[MAXTOKENLEN];
	int errors = 0;

	zero_ups_struct(ups);

	if ((apcconf = fopen(APCCONF, "r")) == NULL) return(1);

	while (fgets(line, sizeof(line), apcconf) != (char *)0) {
		/*
		 * Strip the trailing '\n'
		 *
		 * -RF
		 */
		line[strlen(line)-1] = '\0';

		if (ParseConfig(ups, line) == 0) {
			continue;
		} else {
			errors++;
		}
	}
	fclose(apcconf);

	/*
	 * post-process the configuration stored in the ups structure
	 */

	if (ups->class.type != NETSLAVE) ups->usermagic[0] = '\0';

	if (ups->annoy >= ups->delay) ups->delay = 0;

	if ((ups->sharenet.type == SHAREMASTER) ||
	    (ups->sharenet.type == SHARENETMASTER)) {
		ups->maxtime = 0;
		ups->percent = 10;
		ups->runtime = 5;
	}

	if ((ups->cable.type < CUSTOM_SMART) && ups->mode.type >= BKPRO) {
		fprintf(stderr, "%s: Error :: Changing UPSTYPE from %s ",
		        ups->argvalue, ups->mode.long_name);

		errors += match_range(ups, WHERE(mode), types, "backups");

		fprintf(stderr, "to %s due wrong Cable of Smart Signals.\n\a",
		        ups->mode.long_name);
	}

	if (ups->control_path[0]) {

	sprintf(ups->annoyme,         "%s ANNOYME",    ups->control_path);
	sprintf(ups->changeme,        "%s CHANGEME",   ups->control_path);
	sprintf(ups->doshutdown,      "%s DOSHUTDOWN", ups->control_path);
	sprintf(ups->emergency,       "%s EMERGENCY",  ups->control_path);
	sprintf(ups->failing,         "%s FAILING",    ups->control_path);
	sprintf(ups->loadlimit,       "%s LOADLIMIT",  ups->control_path);
	sprintf(ups->mainsback,       "%s MAINSBACK",  ups->control_path);
	sprintf(ups->onbatt,          "%s ONBATTERY",  ups->control_path);
	sprintf(ups->powerout,        "%s POWEROUT",   ups->control_path);
	sprintf(ups->remote_shutdown, "%s REMOTE",     ups->control_path);
	sprintf(ups->restartme,       "%s RESTARTME",  ups->control_path);
	sprintf(ups->runlimit,        "%s RUNLIMIT",   ups->control_path);
	sprintf(ups->timeout,         "%s TIMEOUT",    ups->control_path);
	sprintf(ups->waitasec,        "%s WAITASEC",   ups->control_path);

	}

	/* JHNC-FIXME: check the logic--what is this supposed to do? */
{
	int len = strlen(ups->lockpath);

	if (len-- > 0) {
		if (ups->lockpath[len] == '/')
			ups->lockpath[len] = 0;
	}
	if (!errors && (ups->cable.type != APC_NET)) {
		char *dev = strrchr(ups->device, '/');

		strcat(ups->lockpath, LOCK_PREFIX);
		strcat(ups->lockpath, dev?++dev:ups->device);
	} else {
		ups->lockpath[0] = 0;
		ups->lockfile = -1;
	}
}

	if ((ups->slave_count > 0) && ups->master_name[0]) {
		fprintf(stderr, "I can't be both MASTER and SLAVE\n");
		errors++;
	}

	switch (ups->nologin.type) {
		case TIMEOUT:
			if (ups->maxtime != 0)
				ups->nologin_time = (int)(ups->maxtime * 0.9);
			break;
		case PERCENT:
			ups->nologin_time = (int)(ups->percent * 1.1);
			if (ups->nologin_time == ups->percent)
				ups->nologin_time++;
			break;
		case MINUTES:
			ups->nologin_time = (int)(ups->runtime * 1.1);
			if (ups->nologin_time == ups->runtime)
				ups->nologin_time++;
			break;
		default:
			break;
	}

	return(errors?1:0);
}

