/*
 *  apcreports.c --
 *
 *  apcupsd.c    -- Simple Daemon to catch power failure signals from a
 *                  BackUPS, BackUPS Pro, or SmartUPS (from APCC).
 *               -- Now SmartMode support for SmartUPS and BackUPS Pro.
 *
 *  Copyright (C) 1996-98 Andre M. Hedrick <hedrick@astro.dyer.vanderbilt.edu>
 *                        Vanderbilt University, Dyer Observatory.
 *  All rights reserved.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <apc_defines.h>
#include <apc_struct.h>
#include <apc_extern.h>

extern FILE *procfile;
extern FILE *logfile;

extern time_t last_time_on_line;
extern time_t last_time_annoy;
extern time_t last_time_delay;
extern time_t last_time_procfs;
extern time_t last_time_logging;
extern time_t last_time_net;

extern int flags;
extern int killcount;

extern int le_bit;
extern int dtr_bit;
extern int rts_bit;
extern int st_bit;
extern int sr_bit;
extern int cts_bit;
extern int cd_bit;
extern int rng_bit;
extern int dsr_bit;

/*********************************************************************/
int updating_procfs(FILE *file, UPSINFO *ups)
{ char msg[100];
  time_t nowtime;

  file = freopen(UPS_STAT, "w", file);
  time(&nowtime);
  strftime(msg, 100, "%b %d %X", localtime(&nowtime));

  fprintf(file,"APC      : %s\n", msg);
  fprintf(file,"CABLE    : %s\n", ups->cable.long_name);
  fprintf(file,"UPSMODEL : %s\n", ups->mode.long_name);
  fprintf(file,"UPSMODE  : %s\n", ups->class.long_name);
  if (ups->sharenet.type != DISABLE)
    fprintf(file,"SHARE    : %s\n", ups->sharenet.long_name);

  switch(ups->mode.type)
    { case BK:
      case SHAREBASIC:
      case NETUPS:
        if (ups->LineUp == 0)
            { fprintf(file, "LINEFAIL : OK\n");
              fprintf(file, "BATTSTAT : OK\n"); }
          else
            { fprintf(file, "LINEFAIL : DOWN\n");
              if ((ups->BattUp == 0) ||
                  ((ups->BattUp != 0) &&
                   ((flags & dtr_bit) == 0)))
                  fprintf(file, "BATTSTAT : RUNNING\n");
                else
                  fprintf(file, "BATTSTAT : FAILING\n"); }
        break;
      case BKPRO:
      case VS:
      case NBKPRO:
      case SMART:
      case SHARESMART:
      case MATRIX:

        if (ups->UPSlink_flags == 0x0000800f)
          { if (ups->LineUp == 0)
                { fprintf(file, "LINEFAIL : OK\n");
                  fprintf(file, "BATTSTAT : OK\n");
                    if (ups->LineLevel == -1)
                      fprintf(file, "LINEVOLT : LOW\n");
                    else if (ups->LineLevel == 1)
                      fprintf(file, "LINEVOLT : HIGH\n");
                    else
                      fprintf(file, "LINEVOLT : OK\n"); }
              else
                { fprintf(file, "LINEFAIL : DOWN\n");
                  if (ups->BattUp == 0)
                      fprintf(file, "BATTSTAT : RUNNING\n");
                    else
                      fprintf(file, "BATTSTAT : FAILING\n"); } }

        if (ups->UPSlink_flags & 0x01000000)
          fprintf(file, "UPSNAME  : %s\n", ups->name);

        if (ups->UPSlink_flags & 0x00000010)
          fprintf(file, "ULINE    : %05.1f Volts\n", ups->LineVoltage);

        if (ups->UPSlink_flags & 0x00000020)
          fprintf(file, "MLINE    : %05.1f Volts\n", ups->LineMax);

        if (ups->UPSlink_flags & 0x00000040)
          fprintf(file, "NLINE    : %05.1f Volts\n", ups->LineMin);

        if (ups->UPSlink_flags & 0x00000800)
          fprintf(file, "FLINE    : %03.1f Hz\n", ups->LineFreq);

        if (ups->UPSlink_flags & 0x00000080)
          fprintf(file, "VOUTP    : %05.1f Volts\n", ups->OutputVoltage);

        if (ups->UPSlink_flags & 0x00000400)
          fprintf(file, "LOUTP    : %05.1f Load Capacity\n", ups->UPSLoad);

        if (ups->UPSlink_flags & 0x00000200)
          fprintf(file, "BOUTP    : %04.1f Volts\n", ups->BattVoltage);

        if (ups->UPSlink_flags & 0x00000100)
          fprintf(file, "BCHAR    : %05.1f Batt. Charge\n", ups->BatLoad);

        if (ups->UPSlink_flags & 0x00001000)
          fprintf(file, "TIME     : %04.1f Minutes\n", ups->TimeLeft);

        if (ups->UPSlink_flags & 0x00010000)
          { switch((*ups).sensitivity[0])
              { case 'L': fprintf(file, "SENSE    : LOW\n");break;
                case 'M': fprintf(file, "SENSE    : MEDIUM\n");break;
                case 'H':
                default : fprintf(file, "SENSE    : HIGH\n");break; } }

        if (ups->UPSlink_flags & 0x00020000)
          fprintf(file, "WAKEUP   : %03d Cycles\n", ups->wakeup);

        if (ups->UPSlink_flags & 0x00040000)
          fprintf(file, "SLEEP    : %03d Cycles\n", ups->sleep);

        if (ups->UPSlink_flags & 0x00080000)
          fprintf(file, "LOTRANS  : %03d.0 Volts\n", ups->lotrans);

        if (ups->UPSlink_flags & 0x00100000)
          fprintf(file, "HITRANS  : %03d.0 Volts\n", ups->hitrans);

        if (ups->UPSlink_flags & 0x00200000)
          fprintf(file, "CHARGE   : %03d.0 Percent\n", ups->minon);

        if (ups->UPSlink_flags & 0x00000001)
          fprintf(file, "BFAIL    : 0x%02X Status Flag\n", ups->Status);

        if (ups->UPSlink_flags & 0x00002000)
          fprintf(file, "UTEMP    : %04.1f C Internal\n", ups->UPSTemp);

        if (ups->UPSlink_flags & 0x00400000)
          { switch((*ups).beepstate[0])
              { case 'T': fprintf(file, "ALARM    : Low Batt + 30\n");break;
                case 'L': fprintf(file, "ALARM    : Low Batt\n");break;
                case 'N': fprintf(file, "ALARM    : Never\n");break;
                case '0':
                default : fprintf(file, "ALARM    : Always\n");break; } }

        if (ups->UPSlink_flags & 0x00000004)
          { switch((*ups).G[0])
              { case 'O':
                  fprintf(file, "LASTEVNT : POWER UP\n");
                  break;
                case 'S':
                  fprintf(file, "LASTEVNT : SELF TEST\n");
                  break;
                case 'L':
                  fprintf(file, "LASTEVNT : LINE VOLTAGE DECREASE\n");
                  break;
                case 'H':
                  fprintf(file, "LASTEVNT : LINE VOLTAGE INCREASE\n");
                  break;
                case 'T':
                  fprintf(file, "LASTEVNT : POWER FAILURE\n");
                  break;
                default :
                  fprintf(file, "LASTEVNT : UNKNOWN EVENT %c\n", (*ups).G[1]);
                  break; } }

        if (ups->UPSlink_flags & 0x00800000)
          fprintf(file, "LOWBATT  : %02d Minutes\n", ups->lowbatt);

        if (ups->UPSlink_flags & 0x00004000)
          { if (ups->seven & 0x08) fprintf(file, "DIPSW    : 0x%d", TRUE);
              else fprintf(file, "DIPSW    : 0x%d", FALSE);
            if (ups->seven & 0x04) fprintf(file, "%d", TRUE);
              else fprintf(file, "%d", FALSE);
            if (ups->seven & 0x02) fprintf(file, "%d", TRUE);
              else fprintf(file, "%d", FALSE);
            if (ups->seven & 0x01)fprintf(file, "%d\n", TRUE);
              else fprintf(file, "%d\n", FALSE); }

        break;
      default: break; }
  fflush(file);
  time(&last_time_procfs);
  return (0); }

/*********************************************************************/
int updating_logs(FILE *file, UPSINFO *ups)
{ char msg[100];
  time_t nowtime;

  time(&nowtime);
  strftime(msg, 100,"%b %d %X", localtime(&nowtime));

  fprintf(file, "APC: %s ", msg);

  switch(ups->mode.type)
    { case BK:
      case SHAREBASIC:
      case NETUPS:
        if (ups->LineUp == 0)
            { fprintf(file, "LINEFAIL:OK ");
              fprintf(file, "BATTSTAT:OK\n"); }
          else
            { fprintf(file, "LINEFAIL:DOWN ");
              if ((ups->BattUp == 0) ||
                  ((ups->BattUp != 0) &&
                   ((flags & dtr_bit) == 0)))
                  fprintf(file, "BATTSTAT:RUNNING\n");
                else
                  fprintf(file, "BATTSTAT:FAILING\n"); }
        break;
      case BKPRO:
      case VS:
      case NBKPRO:
      case SMART:
      case SHARESMART:
      case MATRIX:

        if (ups->UPSlink_flags == 0x0000800f)
            { if (ups->LineUp == 0)
                  { fprintf(file, "LINEFAIL:OK ");
                    fprintf(file, "BATTSTAT:OK ");
                    if (ups->LineLevel == -1)
                        fprintf(file, "LINEVOLT:LOW ");
                      else if (ups->LineLevel == 1)
                        fprintf(file, "LINEVOLT:HIGH ");
                      else
                        fprintf(file, "LINEVOLT:OK "); }
                else
                  { fprintf(file, "LINEFAIL:DOWN ");
                    if (ups->BattUp == 0)
                        fprintf(file, "BATTSTAT:RUNNING ");
                      else
                        fprintf(file, "BATTSTAT:FAILING "); }

              switch ((*ups).G[0])
                { case 'O':
                    fprintf(file, "LASTEVNT:POWER UP\n");
                    break;
                  case 'S':
                    fprintf(file, "LASTEVNT:SELF TEST\n");
                    break;
                  case 'L':
                    fprintf(file, "LASTEVNT:LINE VOLTAGE DECREASE\n");
                    break;
                  case 'H':
                    fprintf(file, "LASTEVNT:LINE VOLTAGE INCREASE\n");
                    break;
                  case 'T':
                    fprintf(file, "LASTEVNT:POWER FAILURE\n");
                    break;
                  default :
                    fprintf(file, "LASTEVNT:UNKNOWN EVENT %c\n", (*ups).G[1]);
                    break; } }
          else
            { fprintf(file, "%05.1f ", ups->LineVoltage);
              fprintf(file, "%05.1f ", ups->LineMax);
              fprintf(file, "%05.1f ", ups->LineMin);
              fprintf(file, "%03.1f ", ups->LineFreq);
              fprintf(file, "%05.1f ", ups->OutputVoltage);
              fprintf(file, "%05.1f ", ups->UPSLoad);
              fprintf(file, "%04.1f ", ups->BattVoltage);
              fprintf(file, "%05.1f ", ups->BatLoad);
              if (ups->UPSlink_flags & 0x00002000)
                fprintf(file, "%04.1f C", ups->UPSTemp);
              fprintf(file, "\n"); }
        break;
      default: break; }
  fflush(file);
  time(&last_time_logging);
  return (0); }

/*********************************************************************/
int logprintf(char *fmt, ...)
{ va_list ap;
  char timestr[32];
  time_t now;
  int ret;

  if (!logfile) return(0);                     /* No logging enabled */

  va_start(ap, fmt);

  /* All entries starts with a timestamp */
  time(&now);
  strftime(timestr, sizeof(timestr), "%b %d %X", localtime(&now));
  fprintf(logfile, "APC: %s ", timestr);

  /* And now for the real message .. */
  ret = vfprintf(logfile, fmt, ap);
  if (fflush(logfile)) ret = -1;

  va_end(ap);
  return(ret); }

