/*
 *  apcupsd.c -- Simple Daemon to catch power failure signals from a
 *               BackUPS, BackUPS Pro, or SmartUPS (from APCC).
 *            -- Now SmartMode support for SmartUPS and BackUPS Pro.
 *
 *  Copyright (C) 1996-98 Andre M. Hedrick <hedrick@astro.dyer.vanderbilt.edu>
 *                        Vanderbilt University, Dyer Observatory.
 *  All rights reserved.
 *
 *  IN NO EVENT SHALL ANDRE HEDRICK, VANDERBILT UNIVERSITY, OR ANY AND ALL
 *  PERSONS INVOLVED IN THE DEVELOPMENT OF THIS PACKAGE, NOW REFERRED TO AS
 *  "APCUPSD-Team" BE LIABLE TO ANY PARTY FOR  DIRECT, INDIRECT, SPECIAL, 
 *  INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 *  SOFTWARE AND ITS DOCUMENTATION, EVEN IF ANY OR ALL OF THE "APCUPSD-Team"
 *  HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  THE "APCUPSD-Team" SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 *  ON AN "AS IS" BASIS, AND THE "APCUPSD-Team" HAS NO OBLIGATION TO PROVIDE
 *  MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *  THE "APCUPSD-Team" HAS ABSOLUTELY NO CONNECTION WITH THE COMPANY
 *  AMERICAN POWER CONVERSION, "APCC".  THE "APCUPSD-Team" DID NOT AND
 *  HAS NOT SIGNED ANY NON-DISCLOSURE AGREEMENTS WITH "APCC".  ANY AND ALL
 *  OF THE LOOK-A-LIKE ( UPSlink(tm) Language ) WAS DERIVED FROM THE
 *  SOURCES LISTED BELOW.
 *
 *  Parts of the code are from Miquel van Smoorenburg's powerd.c
 *  Other parts are original from Christian Holtje <docwhat@uiuc.edu>
 *  Smart-UPS code from Pavel Korensky's apcd.c, apcd.h, and upsnet.c.
 *  I believe that it is okay to say that this is Public Domain, just
 *  give credit, where credit is due.
 *
 *  Disclaimer:  We make NO claims to this software, and take no
 *               resposibility for it's use/misuse.
 *
 *  Cable for a smarter Back-UPS and SmartUPS from APC only.
 *
 *  Computer Side   |  Description of Cable           |  UPS Side
 *  DB9f  |  DB25f  |                                 |    DB9m
 *   4    |   20    |  DTR (5vcc)             *below  |    n/c
 *   8    |    5    |  CTS (low-batt)         *below  |     5
 *   2    |    3    |  RxD (other line-fail)  *below  |     3
 *   5    |    7    |  Ground (Signal)                |     4
 *   1    |    8    |  CD (line-fail from ups)        |     2
 *   7    |    4    |  RTS (shutdown ups)             |     1
 *  n/c   |    1    |  Frame/Case Gnd (optional)      |     9
 *
 *  The "*below" one needs solder type DB connectors with standard
 *  hoods, two (2) 4.7K ohm 1/4 watt %5, one (1) foot of 3/32" (inch)
 *  shrink wrap, rosin core solder, and three (3) to five (5) feet of
 *  22AWG multi-stranded five (5) conductor cable. 
 *
 *  9/20/96 2:00am
 *
 */

/*
 * Add here an identifer that is unique for tracking changes
 * made by each person other than myself.
 *
 * Christopher J. Reimer <reimer@doe.carleton.ca>       CJR
 * Werner Panocha <wpanocha@t-online.de>                w.p.
 * Brian Schau <bsc@fleggaard.dk>                       BSC
 * Marko Sakari Asplund <Marko.Asplund@cern.ch>         MSA
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <apc_version.h>
#include <apc_defines.h>
#include <apc_struct.h>
#include <apc_extern.h>

#undef __DEBUG

int master_present;
int socketfd;
int newsocketfd;

UPSINFO myUPS;
struct sockaddr_in my_adr;
int masterlen;
struct sockaddr_in master_adr;
struct hostent *mastent;

int server_fifo_fd;
int client_fifo_fd;

FILE *procfile = NULL;  /* initial value set to NULL for terminate */
FILE *logfile = NULL;   /* initial value set to NULL for terminate */
FILE *nlgf = NULL;      /* initial value set to NULL for terminate */

time_t last_time_on_line;
time_t last_time_annoy;
time_t last_time_delay;
time_t last_time_procfs;
time_t last_time_logging;
time_t last_time_net;
time_t last_time_access;
time_t last_time_nologon;

int debug_net = FALSE;

int flags;
int killcount;

int le_bit   = TIOCM_LE;
int dtr_bit  = TIOCM_DTR;
int rts_bit  = TIOCM_RTS;
int st_bit   = TIOCM_ST;
int sr_bit   = TIOCM_SR;
int cts_bit  = TIOCM_CTS;
int cd_bit   = TIOCM_CD;
int rng_bit  = TIOCM_RNG;
int dsr_bit  = TIOCM_DSR;

struct termios oldtio;
struct termios newtio;

/*********************************************************************/
void powerfail(int ok)
{ /*  If apcupsd terminates here, it will never get a chance to
   *  report the event of returning mains-power.
   *  I think apcupsd has no need to force terminate() by itself.
   *  It will receive a SIGTERM from init, when system goes down.
   *  This signal is trapped and will trigger apcupsd's terminate()
   *  function.
   *  (Was already reported/fixed in my 2.7.1 diff,)
   *  (so I do it again now) w.p.

  if (ok == 2) terminate(0); }

   *  Closes procfile and logfile to preserve information.
   */

  if (ok == 2) clear_files();

  /*
   *  The network slaves apcupsd needs to terminate here for now.
   *  This sloppy, but it works. If you are networked, then the
   *  master must fall also. This is required so that the UPS
   *  can reboot the slaves.
   */

  if (ok == 3) terminate(0); }
  
/*********************************************************************/
void logonfail(int ok)
{ int lgnfd;

  unlink(NOLOGIN);
  if (ok == 0 && ((lgnfd = open(NOLOGIN, O_CREAT|O_WRONLY, 0644)) >= 0))
    { write(lgnfd, POWERFAIL "\n", sizeof(POWERFAIL));
      close(lgnfd); } }

/*********************************************************************/
int make_file(const char *s)
{ int makefd;

  if ((makefd = open(s,O_CREAT|O_WRONLY, 0644)) >= 0)
    close(makefd);
  return makefd; }

/*********************************************************************/
int remove_file(const char *s)
{ unlink(s);
  return *s; }

/**********************************************************************
 * the terminate function and trapping signals allows apcupsd
 * to exit and cleanly close logfiles, and reset the tty back
 * to its original settings. You may want to add this
 * to all configurations.  Of course, the file descriptors
 * must be global for this to work, I also set them to
 * NULL initially to allow terminate to determine whether
 * it should close them.
 *********************************************************************/
void terminate(int sig)
{ sleep(10);
  if (logfile != NULL)
    logprintf("apcupsd exiting, signal %u\n", sig);

  clear_files();

  if (myUPS.fd != -1)
    { delete_lockfile(&myUPS, TRUE);            /* BSC */
      tcflush(myUPS.fd, TCIFLUSH);
      tcsetattr(myUPS.fd, TCSANOW, &oldtio);
      close(myUPS.fd); }
  _exit(0); }

/*********************************************************************/
void clear_files(void)
{ if (logfile != NULL)
    { fflush(logfile);
      fclose(logfile); }
  if (procfile != NULL)
    { fflush(procfile);
      fclose(procfile); } }

/*********************************************************************/
void make_pid(void)
{ pid_t pid;
  FILE *pidf;

  if ((pidf = fopen(APCPID, "w")) != NULL)
    { fprintf(pidf, "%d\n", pid = getpid());
      fclose(pidf); } }

/*********************************************************************/
int setup_serial(UPSINFO *ups)
{ char response[32];    /* w.p. */
  char a;
  static int has_done_setup = 0;
  response[0]='\0';
  
  if (!has_done_setup)
      /*
       * Marko Sakari Asplund <Marko.Asplund@cern.ch>
       * 
       * prevents double init of serial port 9/25/98
       */
      { has_done_setup++; }
    else
      { fprintf(stderr, "%s: serial port already initialized.\n",
                ups->argvalue);
        exit(1); }
  
  /* Open the the device */
  if ((ups->fd = open(ups->device ,O_RDWR | O_NOCTTY | O_APPEND)) < 0)
    { fprintf(stderr, "%s: Cannot open UPS tty %s: %s\n",
               ups->argvalue,
               ups->device,
               strerror(errno));
      return(1); }

  /* next 5 lines by BSC */
  if (create_lockfile(ups))
    { close(ups->fd);
      ups->fd = -1;
      delete_lockfile(ups, FALSE);
      return(1); }

  tcgetattr(ups->fd, &oldtio); /* Save old settings */

  signal(SIGHUP, terminate);
  signal(SIGINT, terminate);
  signal(SIGTERM, terminate);

  newtio.c_cflag = DEFAULT_SPEED | CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR;             /* Ignore errors, raw input */
  newtio.c_oflag = 0;                  /* Raw output */
  newtio.c_lflag = 0;                  /* No local echo */

#ifdef __freebsd__
  newtio.c_ispeed = DEFAULT_SPEED;     /* Set input speed */
  newtio.c_ospeed = DEFAULT_SPEED;     /* Set output speed */
#endif /* __freebsd__ */

  /* w.p. This makes a non.blocking read() with 5 sec. timeout */
  newtio.c_cc[VMIN] = 0;
  newtio.c_cc[VTIME] = 50;

  tcflush(ups->fd, TCIFLUSH);
  tcsetattr(ups->fd, TCSANOW, &newtio);

  /* Have to clear RTS line to access the serial cable mode */
  if ((ups->cable.type == APC_940_0095A) || (ups->cable.type == APC_940_0095C))
    (void) ioctl(ups->fd, TIOCMBIC, &rts_bit);

  if ((ups->cable.type >= CUSTOM_SMART) && (ups->mode.type >= BKPRO))
    { a = 'Y';                      /* Verify that we can talk to UPS w.p. */
      write(ups->fd, &a, 1);        /* This one might not work, if UPS is */
      sleep(1);                     /* in an unstable communication state */
      tcflush(ups->fd, TCIOFLUSH);  /* Discard UPS's response, if any */

      a = 'Y';                      /* Now we should be synchronized .. */
      write(ups->fd, &a, 1);
      getline(response, ups); 

      if (strcmp("SM", response))   /* w.p. */
        { fprintf(stderr,
                  "%s: PANIC! Cannot talk to UPS\n\a",
                  ups->argvalue);
          sleep(2);
          if ((ups->mode.type >= BKPRO) &&
              ((ups->lockfile = open(ups->lockpath, O_CREAT|O_EXCL|O_RDWR, 0644)) < 0))
            { close(ups->lockfile);
              delete_lockfile(ups, TRUE); }
          terminate(0); } }
  return(0); }

/*********************************************************************/
/* this function by BSC */
int create_lockfile(UPSINFO *ups)
{ char pidbuffer[12];
  int pid = 0;
  int attempts = 0;

  if (ups->lockpath)
    { while (TRUE)
        { if (attempts > 5)
            { fprintf(stderr,
                      "%s: can't create lockfile ... giving up!\n",
                      ups->argvalue);
              delete_lockfile(ups, FALSE);
              return(1); }
          if ((ups->lockfile = open(ups->lockpath, O_CREAT|O_EXCL|O_RDWR, 0644)) < 0)
              { if (errno == EEXIST)
                    { if ((pid = read(ups->lockfile, &pidbuffer, 11)) != -1)
                        { pid = 0;
                          close(ups->lockfile);
                          ups->lockfile = -1;
                          pid=sscanf(pidbuffer, " %d", &pid);
                          if ((pid == getpid()) ||
                              (kill(pid, 0) == -1 && errno == ESRCH) ||
                               (pid == 0))
                            { if (unlink(ups->lockpath)) break;
                              attempts++;
                              continue; } }
                      fprintf(stderr, "%s: device locked (%s)\n",
                              ups->argvalue, ups->device);
                      return(1); }
                  else break; }
            else
              { sprintf(pidbuffer, "%010d", getpid());
                write(ups->lockfile, pidbuffer, 11);
                close(ups->lockfile);
                ups->lockfile = -1;
                return(0); } }
    fprintf(stderr, "%s: %s: %s\n", ups->argvalue, ups->lockpath,
            strerror(errno));
    delete_lockfile(ups, FALSE);
    return(1); }
  return(0); }

/*********************************************************************/
/* this function by BSC */
void delete_lockfile(UPSINFO *ups, int boolean)
{ if (ups->lockpath)
    { if (boolean)
        { unlink(ups->lockpath);
          ups->lockpath[0] = '\0'; }

      if (ups->lockfile != -1)
        { close(ups->lockfile);
          ups->lockfile = -1; } } }

/*********************************************************************/
/* this function by BSC */
void rethink_lockfile(UPSINFO *ups)
{ char pidbuffer[12];

  if (ups->lockpath)
    { if ((ups->lockfile = open(ups->lockpath, O_CREAT|O_RDWR, 0644)) < 0)
        { perror(ups->argvalue);
          terminate(0); }
      sprintf(pidbuffer, "%010d", getpid());
      write(ups->lockfile, pidbuffer, 11);
      close(ups->lockfile);
      ups->lockfile = -1; } }

/*********************************************************************/
void kill_serial(UPSINFO *ups, int killpwr)
{ char response[32];                                         /* w.p. */
  char a;
  FILE *pwdf;
  int errflag = 0;
  response[0] = '\0';

  if (killpwr)
      { if (ups->mode.type <= SHAREBASIC)
          { for (killcount=0;killcount<3;killcount++)
              { ioctl(ups->fd, TIOCMGET, &flags);
                sleep(1);
                switch(ups->cable.type)
                  { case CUSTOM_SIMPLE:
                      ups->LineUp = (flags & cd_bit);
                      break;
                    case APC_940_0020B:
                      ups->LineUp = (flags & cts_bit);
                      break;
                    case APC_940_0023A:
                      ups->LineUp = (flags & cd_bit);
                      break;
                    case APC_940_0095A:
                    case APC_940_0095C:
                      ups->LineUp = (flags & rng_bit);
                      break;
                    case CUSTOM_SMART:
                    case APC_940_0024B:
                    case APC_940_0024C:
                    case APC_NET:
                      break;
                    default:
                      ups->LineUp = 1;
                      break; } } }
        if ((((pwdf = fopen(PWRFAIL, "r" )) == NULL) &&
             (ups->mode.type != BK)) ||
            (((pwdf = fopen(PWRFAIL, "r" )) == NULL) &&
             (ups->LineUp != 0) && (ups->mode.type == BK)))
            { fprintf(stderr,
                      "%s: Attempting to kill the power!\n" \
                      "%s: Nice TRY but not TODAY!\n" \
                      "%s: Shutdown not called first.\n" \
                      "%s: Bug found by Tom Kunicki\n",
                      ups->argvalue, ups->argvalue,
                      ups->argvalue, ups->argvalue);
              terminate(0); }
          else
            { errflag=0;                                     /* w.p. */
              if ((ups->class.type == SHAREMASTER) ||
                  (ups->class.type == SHARENETMASTER))
                { fprintf(stderr,
                          "%s: Waiting 30 seconds for slave(s) to shutdown.\n" \
                          "%s: BETA blind code writing. Here!!\n" \
                          "%s: Someone tell me if this works...\n",
                          ups->argvalue, ups->argvalue, ups->argvalue);
                  sleep(30); }
              if (ups->class.type == NETMASTER)
                { fprintf(stderr,
                          "%s: Waiting 15 seconds for slave(s) to shutdown.\n",
                          ups->argvalue);
                  sleep(15); }
              fclose(pwdf);

              fprintf(stderr,
                      "%s: Attempting to kill the power!\n",
                      ups->argvalue);

              if (ups->mode.type == BK)
                  { switch(ups->cable.type)
                      { case CUSTOM_SIMPLE:           /* killpwr_bit */
                        case APC_940_0095A:
                        case APC_940_0095C:          /* killpwr_bit */
                          (void) ioctl(ups->fd, TIOCMBIS, &rts_bit);
                          (void) ioctl(ups->fd, TIOCMBIS, &rts_bit);
                          (void) ioctl(ups->fd, TIOCMBIS, &rts_bit);
                          break;
                        case APC_940_0020B:           /* killpwr_bit */
                          (void) ioctl(ups->fd, TIOCMBIS, &dtr_bit);
                          (void) ioctl(ups->fd, TIOCMBIS, &dtr_bit);
                          (void) ioctl(ups->fd, TIOCMBIS, &dtr_bit);
                          break;
                        case APC_940_0023A:          /* killpwr_bit */
                          break;
                        case CUSTOM_SMART:
                        case APC_940_0024B:
                        case APC_940_0024C:
                        case APC_NET:
                        default:
                          break; }
                    sleep(10);
                    close(ups->fd);
                    fprintf(stderr,
                            "%s: Drat!! Failed to kill the power!\n" \
                            "%s: Great Mains Returned, Trying a REBOOT!\n",
                            ups->argvalue, ups->argvalue);
                    system("reboot");
                    fprintf(stderr,
                            "%s: Perform CPU-reset or power-off\n",
                            ups->argvalue);
                    terminate(0); }
                else if (ups->mode.type == SHAREBASIC)
                  { sleep(10);
                    close(ups->fd);
                    fprintf(stderr,
                            "%s: Waiting For ShareUPS Master to shutdown.\n",
                            ups->argvalue);
                    sleep(60);
                    fprintf(stderr,
                            "%s: Drat!! Failed to have power killed by Master!\n" \
                            "%s: Great Mains Returned, Trying a REBOOT!\n",
                            ups->argvalue, ups->argvalue);
                    system("reboot");
                    fprintf(stderr,
                            "%s: Perform CPU-reset or power-off\n",
                            ups->argvalue);
                    terminate(0); }
                else
                  { a = 'S';
                    write(ups->fd, &a, 1);
     /*  Check wether the UPS has acknowledged the power-off command.
      *  This might not happen in rare cases, when mains-power returns
      *  just after LINUX starts with the shutdown sequence. 
      *   If the UPS doesnt acknowledge the command, it will not interrupt
      *  the ouput-power. So LINUX will not come up without operator
      *  intervention.  w.p.
      */
                    sleep(10);
                    getline(response, ups);
                    if (strcmp(response, "OK") == 0)
                      fprintf(stderr, "%s: UPS will power off ...\n",
                               ups->argvalue);
                    if (strcmp(response,"NA") == 0)
                      { fprintf(stderr,
                                "%s: Mains-power returned during shutdown!\n" \
                                "%s: Trying alternate method for killing power.\n",
                                ups->argvalue, ups->argvalue);
                        /*
                         *  w.p. experiments shows that UPS needs
                         *  w.p. delays between chars to accept
                         *  w.p. this command
                         */
                        a = '@';
                        write(ups->fd, &a, 1);
                        sleep(1);
                        a = '0';
                        write(ups->fd, &a, 1); 
                        sleep(1);
                        a = '0';
                        write(ups->fd, &a, 1);
                        sleep(1); 
                        a = '0';
                        write(ups->fd, &a, 1);
                        sleep(2);
                        getline(response, ups);
                        if (strcmp(response, "OK") == 0)
                            fprintf(stderr,
                                    "%s: UPS power-cycle in progress ...\n",
                                    ups->argvalue);
                          else
                            errflag++; }
                    if (strcmp(response, "OK") && strcmp(response, "NA"))
                      { fprintf(stderr,
                                "%s: Mains-power returned during shutdown!\n" \
                                "%s: Trying alternate method for killing power.\n" \
                                "%s: Using non SU SmartUPS alternate method.\n",
                                ups->argvalue, ups->argvalue, ups->argvalue);
                        a = '@';
                        write(ups->fd, &a, 1);
                        sleep(1);
                        a = '0';
                        write(ups->fd, &a, 1);
                        sleep(1);
                        a = '0';
                        write(ups->fd, &a, 1);
                        sleep(1);
                        if ((ups->mode.type == BKPRO) || (ups->mode.type == VS))
                            { a = '1';
                              fprintf(stderr,
                                      "%s: BackUPS Pro and SmartUPS v/s sleep for 6 min.\n",
                                      ups->argvalue); }
                          else
                            { a = '0'; }
                        write(ups->fd, &a, 1);
                        sleep(1);
                        a = 'K';
                        write(ups->fd, &a, 1);
                        sleep(2);
                        getline(response, ups);
                        if ((strcmp(response,"*") == 0) ||
					(ups->mode.type >= BKPRO))
                            fprintf(stderr,
                                    "%s: UPS power-cycle in progress ...\n",
                                    ups->argvalue);
                          else
                            errflag++; }
                    if (errflag)
                      { fprintf(stderr,
                                "%s: Unexpected error!\n" \
                                "%s: Perform CPU-reset or power-off\n",
                                ups->argvalue, ups->argvalue);
                        system("reboot"); } }
                    /* w.p. */
              sleep(10);
              terminate(0); } }
    else
      { if (ups->mode.type <= SHAREBASIC)
          { switch(ups->cable.type)
              { case CUSTOM_SIMPLE:
                  (void) ioctl(ups->fd, TIOCMBIC, &dtr_bit);  /* power_bit */
                  (void) ioctl(ups->fd, TIOCMBIC, &rts_bit);  /* killpwr_bit */
                  break;
                case APC_940_0020B:
                  (void) ioctl(ups->fd, TIOCMBIC, &dtr_bit);  /* killpwr_bit */
                  break;
                case APC_940_0023A:
                  break;
                case APC_940_0095A:
                case APC_940_0095C:
                  (void) ioctl(ups->fd, TIOCMBIC, &rts_bit);  /* killpwr_bit */
                  (void) ioctl(ups->fd, TIOCMBIC, &cd_bit);   /* lowbatt_bit */
                  (void) ioctl(ups->fd, TIOCMBIC, &rts_bit);  /* killpwr_bit */
                  break;
                case CUSTOM_SMART:
                case APC_940_0024B:
                case APC_940_0024C:
                case APC_NET:
                default:
                  fprintf(stderr,
                          "%s: Simple Cable Signal Lost!!!\n\a",
                          ups->argvalue);
                  terminate(0);
                  break; } } } }

/*********************************************************************/
int check_serial(UPSINFO *ups)
{ if (ups->mode.type <= SHAREBASIC)
      { ioctl(ups->fd, TIOCMGET, &flags);
        switch(ups->cable.type)
          { case CUSTOM_SIMPLE:
              ups->LineUp = (flags & cd_bit);
              if (flags & cts_bit) ups->BattUp = 0;
                else ups->BattUp = 1;
              ups->LineDown = (flags & sr_bit);
              break;
            case APC_940_0020B:
              ups->LineUp = (flags & cts_bit);
              ups->BattUp = (flags & cd_bit);
              ups->LineDown = 0;
              break;
            case APC_940_0023A:
              ups->LineUp = (flags & cd_bit);

/* BOGUS STUFF MUST FIX IF POSSIBLE */

              ups->BattUp = (flags & sr_bit);

/* BOGUS STUFF MUST FIX IF POSSIBLE */

              ups->LineDown = 0;
              break;
            case APC_940_0095A:
            case APC_940_0095C:
              ups->LineUp = (flags & rng_bit);
              ups->BattUp = (flags & cd_bit);
              ups->LineDown = 0;
              break;
            case CUSTOM_SMART:
            case APC_940_0024B:
            case APC_940_0024C:
            case APC_NET:
            default:
              return(1); } }
    else
      fillUPS(ups);
  sleep(1);
  return(0); }
  
/*********************************************************************/
void do_serial(UPSINFO *ups)
{ int count = 0;
  int i = 0;
  int slave_bool = FALSE;

  rethink_lockfile(ups);

  if (ups->mode.type >= BKPRO)
    setup_extended(ups);

  if (ups->slave_count >= 1)
    slave_bool = TRUE;

  if (logfile != NULL)
    time(&last_time_logging);

  if (procfile != NULL)
    time(&last_time_procfs);

  for (i=0;i<ups->slave_count;i++)
    logprintf("found slave: %s\n", ups->slaves[i].name);

  if (ups->nettime != 0)
    time(&last_time_net);

  while(1)
    { if ((!ups->LineUp) && (ups->maxtime != 0))
        time(&last_time_on_line);
      if (!ups->LineUp) time(&last_time_annoy);
      if (!ups->LineUp) time(&last_time_delay);
      if (!ups->LineUp) time(&last_time_nologon);
      /*
       *  Now sample the status_bit line.
       */
      if (check_serial(ups))
        logprintf("%s: Simple Cable Signal Lost!!!\n", ups->argvalue);

      if (ups->enable_access.type == TRUE)
        { (void) pipe_requests(ups);
          /*
           *  Must SHUTDOWN Remote System Calls
           */
          if (ups->remotedown)
            { ups->BatteryUp = 0;
              ups->ShutDown = 1;
              if (slave_bool) slave_bool = send_to_slaves(ups);
              if (ups->remote_cmd[0]) system(ups->remote_cmd);
              delete_lockfile(ups, TRUE);
              if (ups->nologin_file != TRUE)
                make_file(NOLOGIN);
              ups->nologin_file = TRUE;
              logonfail(0);
              system(ups->remote_shutdown);
              make_file(PWRFAIL);
              powerfail(3); } }
      /*
       *  Did status_bit go high? Then the power is failing.
       */
      if (ups->OldLineUp == 0 && ups->LineUp != 0)
        { count++;
          if (count > 3)
            { system(ups->powerout);
              if (slave_bool) slave_bool = send_to_slaves(ups);
              if (logfile != NULL) updating_logs(logfile, ups);
              if (procfile != NULL) updating_procfs(procfile, ups);
              if (ups->pwr_cmd[0]) system(ups->pwr_cmd);
              /*
               *  Now enable the DTR. (power_bit) for _CUSTOM
               */
              if (ups->cable.type == CUSTOM_SIMPLE)
                (void) ioctl(ups->fd, TIOCMBIS, &dtr_bit);
              ups->OldLineUp = ups->LineUp;
              ups->OldBattUp = ups->BattUp;
              ups->OldLineDown = ups->LineDown;
              sleep(10);
              /*
               *  Now re-sample the status_bit line.
               */
              (void) check_serial(ups);
              sleep(1);
              /*
               *  Did the second test verify the power is failing?
               */
              if ((ups->LineUp) && (ups->OldLineUp))
                { if (slave_bool) slave_bool = send_to_slaves(ups);
                  if (logfile != NULL) updating_logs(logfile, ups);
                  if (procfile != NULL) updating_procfs(procfile, ups);
                  ups->BatteryUp = 1;
/*
                  if (ups->nologin.type == ALWAYS)
                    { make_file(NOLOGIN);
                      ups->nologin_file = TRUE;
                      logonfail(0); }
*/
                  system(ups->onbatt); } }
          else continue; }
      /*
       *  Did lowbatt_bit go high? Then the battery power is failing.
       *  Normal Power down during Power Failure
       */
      if ((ups->OldBattUp == 0 && ups->BattUp > 0) &&
          (ups->BatteryUp == 1 && ups->ShutDown == 0))
        { count++;
          if (count > 3)
              { ups->BatteryUp = 0;
                ups->ShutDown = 1;
                if (slave_bool) slave_bool = send_to_slaves(ups);
                delete_lockfile(ups, TRUE);
                system(ups->failing);
                if (ups->batt_cmd[0]) system(ups->batt_cmd);
                make_file(PWRFAIL);
                powerfail(2);
                if (ups->nologin_file != TRUE)
                  make_file(NOLOGIN);
                logonfail(0);
                system(ups->doshutdown); }
            else continue; }
      /*
       *  Did MaxTimeOnBattery Expire?
       *  Normal Power down during Power Failure
       */
      if ((ups->maxtime != 0) && ups->LineUp &&
          ((time(NULL) - last_time_on_line) > ups->maxtime))
        { count++;
          if (count > 3)
              { ups->ShutDown = 1;
                ups->timedout = TRUE;
                if (slave_bool) slave_bool = send_to_slaves(ups);
                if (logfile != NULL) updating_logs(logfile, ups);
                if (procfile != NULL) updating_procfs(procfile, ups);
                if (ups->time_cmd[0]) system(ups->time_cmd);
                delete_lockfile(ups, TRUE);
                system(ups->timeout);
                make_file(PWRFAIL);
                powerfail(2);
                if (ups->nologin_file != TRUE)
                  make_file(NOLOGIN);
                logonfail(0);
                system(ups->doshutdown); }
            else continue; }
      /*
       *  Did Battery Charge go below percent cutoff?
       *  Normal Power down during Power Failure
       */
      if ((ups->mode.type >=  NBKPRO) && ups->LineUp &&
          (ups->BatLoad <= ups->percent))
        { count++;
          if (count > 3)
              { ups->ShutDown = 1;
                ups->load = TRUE;
                if (slave_bool) slave_bool = send_to_slaves(ups);
                if (logfile != NULL) updating_logs(logfile, ups);
                if (procfile != NULL) updating_procfs(procfile, ups);
                if (ups->load_cmd[0]) system(ups->load_cmd);
                delete_lockfile(ups, TRUE);
                system(ups->loadlimit);
                make_file(PWRFAIL);
                powerfail(2);
                if (ups->nologin_file != TRUE)
                  make_file(NOLOGIN);
                logonfail(0);
                system(ups->doshutdown); }
            else continue; }
      /*
       *  Did Battery Runtime go below percent cutoff?
       *  Normal Power down during Power Failure
       */
      if ((ups->mode.type >=  NBKPRO) && ups->LineUp &&
          (ups->TimeLeft <= ups->runtime))
        { count++;
          if (count > 3)
              { ups->ShutDown = 1;
                ups->timelout = TRUE;
                if (slave_bool) slave_bool = send_to_slaves(ups);
                if (logfile != NULL) updating_logs(logfile, ups);
                if (procfile != NULL) updating_procfs(procfile, ups);
                if (ups->limit_cmd[0]) system(ups->limit_cmd);
                delete_lockfile(ups, TRUE);
                system(ups->runlimit);
                make_file(PWRFAIL);
                powerfail(2);
                if (ups->nologin_file != TRUE)
                  make_file(NOLOGIN);
                logonfail(0);
                system(ups->doshutdown); }
            else continue; }
      /*
       *  Did status_bit go down again? Then the power is back.
       *  No shutdown to cancel.
       */
      if ((ups->OldLineUp > 0 && ups->LineUp == 0) && ups->ShutDown == 0)
        { count++;
          if (count > 3)
              { ups->BatteryUp = 0;
                system(ups->mainsback);
                if (slave_bool) slave_bool = send_to_slaves(ups);
                if (logfile != NULL) updating_logs(logfile, ups);
                if (procfile != NULL) updating_procfs(procfile, ups);
                if (ups->ret_cmd[0]) system(ups->ret_cmd);
                logonfail(1);
                remove_file(NOLOGIN);
                ups->nologin_file = FALSE;
                if (ups->cable.type == CUSTOM_SIMPLE)
                  { (void) ioctl(ups->fd, TIOCMBIC, &dtr_bit);
                        /*
                         *  Clearing DTR and TxD (ST).
                         */
                    (void) ioctl(ups->fd, TIOCMBIC, &st_bit); }  }
            else continue; }
      /*
       *  Did status_bit go down again? Then the power is back.
       *  Shutdown to cancel.
       */
      if ((ups->OldLineUp > 0 && ups->LineUp == 0) && ups->ShutDown == 1)
        { count++;
          if (count > 3)
              { ups->BatteryUp = 0;
                ups->ShutDown = 0;
                system(ups->waitasec);
                if (slave_bool) slave_bool = send_to_slaves(ups);
                if (logfile != NULL) updating_logs(logfile, ups);
                if (procfile != NULL) updating_procfs(procfile, ups);
                if (ups->ret_cmd[0]) system(ups->ret_cmd);
                powerfail(1);
                remove_file(PWRFAIL);
                logonfail(1);
                remove_file(NOLOGIN);
                ups->nologin_file = FALSE;

                if (ups->cable.type == CUSTOM_SIMPLE)
                  { (void) ioctl(ups->fd, TIOCMBIC, &dtr_bit);
                        /*
                         *  Clearing DTR and TxD (ST).
                         */
                    (void) ioctl(ups->fd, TIOCMBIC, &st_bit); } }
            else continue; }
      /*
       *  Announce to LogOff, with initial delay ....
       */
      if ((ups->BatteryUp == 1 && ups->ShutDown == 0 && ups->LineUp > 0) &&
          ((time(NULL) - last_time_delay) > ups->delay) &&
          ((time(NULL) - last_time_annoy) > ups->annoy) &&
          (ups->nologin_file == TRUE))
        { system(ups->annoyme);
          if (slave_bool) slave_bool = send_to_slaves(ups);
          if (ups->pwr_cmd[0]) system(ups->pwr_cmd);
          time(&last_time_annoy); }

      /*
       *  Delay NoLogons....
       */
      if ((ups->BatteryUp == 1) && (ups->ShutDown == 0) &&
          (ups->LineUp > 0) && (ups->nologin_file != TRUE))
        { count++;
          if (count > 3)
            { switch(ups->nologin.type)
                { case NEVER:
                    break;
                  case TIMEOUT:
                    if ((time(NULL) - last_time_nologon) > ups->nologin_time)
                      { make_file(NOLOGIN);
                        ups->nologin_file = TRUE;
                        logonfail(0); }
                    break;
                  case PERCENT:
                    if (ups->nologin_time >= ups->BatLoad)
                      { make_file(NOLOGIN);
                        ups->nologin_file = TRUE;
                        logonfail(0); }
                    break;
                  case MINUTES:
                    if (ups->nologin_time >= ups->TimeLeft)
                      { make_file(NOLOGIN);
                        ups->nologin_file = TRUE;
                        logonfail(0); }
                    break;
                  case ALWAYS:
                  default:
                    make_file(NOLOGIN);
                    ups->nologin_file = TRUE;
                    logonfail(0);
                    break; } }
            else continue; }
      /*
       *  Did lowbatt_bit never go low then high?
       *  Then the battery power has failed!!!
       *  Must SHUTDOWN NOW
       */
      if ((ups->LineUp != 0 && ups->BattUp == 1 && ups->LineDown == 0) &&
          (ups->BatteryUp == 1 && ups->ShutDown == 0))
        { count++;
          if (count > 3)
              { ups->BatteryUp = 0;
                ups->ShutDown = 1;
                ups->emergencydown = TRUE;
                if (slave_bool) slave_bool = send_to_slaves(ups);
                delete_lockfile(ups, TRUE);
                system(ups->emergency);
                if (ups->batt_cmd[0]) system(ups->batt_cmd);
                if (ups->nologin_file != TRUE)
                  make_file(NOLOGIN);
                ups->nologin_file = TRUE;
                logonfail(0);
                make_file(PWRFAIL);
                powerfail(2); }
            else continue; }

      if ((logfile != NULL) &&
          ((time(NULL) - last_time_logging) > ups->logtime))
        { updating_logs(logfile, ups); }             /* Do Logging */

      if ((procfile != NULL) &&
          ((time(NULL) - last_time_procfs) > ups->proctime))
        { updating_procfs(procfile, ups); }          /* Do Procfs */

      if ((slave_bool) &&
          ((time(NULL) - last_time_net) > ups->nettime))
        { slave_bool = send_to_slaves(ups); }        /* Do slaves */

      /*
       *  Reset count, remember status and sleep 2 seconds.
       */
      count = 0;
      ups->OldLineUp = ups->LineUp;
      ups->OldBattUp = ups->BattUp;
      ups->OldLineDown = ups->LineDown;
      sleep(2); } }

/*********************************************************************/
/*                     Main program.                                 */
/*********************************************************************/
int main(int argc, char **argv)
{
  int KILL = FALSE;
 
  if ((argc >= 2) && ((strcmp(argv[1], "-v") == 0) ||
                      (strcmp(argv[1], "--version") == 0)))
      { fprintf(stderr, "%s v%s Andre Hedrick et al.\n",
                argv[0], APCUPSD_RELEASE);
        fprintf(stderr, "Usage: %s <option>\n\a", argv[0]);
        fprintf(stderr, "NOTE : %s ONLY one option at a time!!!\n\n", argv[0]);
        fprintf(stderr, "       killpower      :\n");
        fprintf(stderr, "       -d | --debug   :\n");
        fprintf(stderr, "       -v | --version :\n");
        exit(1); }
    else if (argc >= 2  && ((strcmp(argv[1], "-d") == 0) ||
                            (strcmp(argv[1], "--debug") == 0)))
      { debug_net = TRUE;

#ifdef DEBUG
        fprintf(stderr, "%s: Built on %s at %s\n", argv[0], __DATE__, __TIME__);
#endif

        sprintf(myUPS.argvalue, "%s", argv[0]);
        KILL = FALSE;
        if (check_for_config(&myUPS)) exit(1);
          else delete_lockfile(&myUPS, KILL);
        if ((myUPS.class.type == NETSLAVE) || (myUPS.class.type == NETMASTER))
            fprintf(stderr, "%s Net Debug mode enabled.\n\a", argv[0]);
          else
            { fprintf(stderr, "%s Network mode not enabled, bye.\n\a",
                      argv[0]);
              exit(1); }
        if (myUPS.logtime == 0)
          { myUPS.logtime = 500;
            fprintf(stderr, "%s Net Debug requires a logtime--newtime: %d.\n\a",
                    argv[0], myUPS.logtime); } }
    else if (argc >= 2  && (strcmp(argv[1], "killpower") == 0))
      { sprintf(myUPS.argvalue, "%s", argv[0]);
        KILL = TRUE;
        if (check_for_config(&myUPS)) exit(1);
          else delete_lockfile(&myUPS, KILL); }
    else if (argc >= 2)
      { fprintf(stderr, "Usage: %s <option>\n\a", argv[0]);
        fprintf(stderr, "NOTE : %s ONLY one option at a time!!!\n\n", argv[0]);
        fprintf(stderr, "       killpower      :\n");
        fprintf(stderr, "       -d | --debug   :\n");
        fprintf(stderr, "       -v | --version :\n");
        exit(1); }
    else
      { sprintf(myUPS.argvalue, "%s", argv[0]);
        KILL = FALSE;
        if (check_for_config(&myUPS)) exit(1);
          else delete_lockfile(&myUPS, KILL); }

  if (!((myUPS.slave_count > 0) && (KILL == FALSE)))
      { debug_net = FALSE; }

  if (myUPS.logtime != 0)
      logfile = fopen(LOG_FILE,"a+");
    else unlink(LOG_FILE);

  if (myUPS.proctime != 0)
      procfile = fopen(UPS_STAT,"w");
    else unlink(UPS_STAT);

  master_present = 0;

  switch(myUPS.sharenet.type)
    { case DISABLE:
      case SHARE:
        if (setup_serial(&myUPS))
          exit(1);
	break;          /* MSA the missing break error......core dump fix? */
      case NET:
        switch(myUPS.class.type)
          { case NO_CLASS:
            case STANDALONE:
            case SHARESLAVE:
            case SHAREMASTER:
            case SHARENETMASTER:
              break;
            case NETSLAVE:
              if ((KILL == 1) || (prepare_slave(&myUPS)))
                exit(1);
              break;
            case NETMASTER:
              if (setup_serial(&myUPS))
                exit(1);
              if ((KILL == 0) && (prepare_master(&myUPS)))
                exit(1);
              break;
            default:
              fprintf(stderr, "%s: NET Class Error %s\n\a",
                      argv[0], strerror(errno));
              exit(1); }
        break;
      case SHARENET:
        if (setup_serial(&myUPS)) exit(1);
        if ((KILL == 0) && (prepare_master(&myUPS))) exit(1);
        break;
      default:
        exit(1); }

  if (myUPS.fd != -1) kill_serial(&myUPS, KILL);
    else kill_net(&myUPS, KILL);

   /*
    *  Become a daemon.
    */
  switch(fork())
    { case 0:                       /* I am the child. */
        setsid();
        break;
      case -1:                      /* Failed to become daemon. */
        fprintf(stderr, "%s: can't fork.\n", argv[0]);
        exit(1);
      default:                     /* I am the parent. */
        exit(0); }

  make_pid();

  if (myUPS.fd != -1) do_serial(&myUPS);
    else do_net(&myUPS);

  /*
   *  Error! (shouldn't happen)
   */
  return(1); }
